#  
#   Copyright (C) 2006-2007 Brian Elliott Finley
#
#   $Id: SystemStateManager.pm 185 2007-09-19 20:52:28Z finley $
#    vi: set filetype=perl:
# 

package SystemStateManager;

use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(version user_is_root get_current_state check_in read_definition_file sync_state add_file_to_repo);

use strict;
use File::Copy;
use File::Path;
use File::Basename;
use Unix::Mknod qw(:all);
use File::stat qw(:FIELDS);
#
# The POSIX library (with qw(mkfifo)) is exporting the S_IS* functions, so we
# need to limit which functions Fcntl exports by explicitly listing the ones
# we need. -BEF-
use POSIX qw(mkfifo);
use Fcntl qw( S_IFBLK S_IFCHR );
use Digest::MD5;


################################################################################
#
#   This package provides the following functions
#
#   add_file_to_repo
#   _backup
#   check_in
#   check_depends
#   choose_tmp_file
#   diff_file
#   do_hardlink
#   do_regular_file
#   do_directory
#   do_softlink
#   do_special_file
#   do_unwanted_file
#   do_you_want_me_to
#   _get_arch
#   get_deb_pkg_dependencies
#   get_deb_pkg_reverse_dependencies
#   get_deb_pkgs_currently_installed
#   get_deb_pkgs_that_need_to_be_upgraded
#   get_file
#   get_pkg_dependencies
#   get_pkg_reverse_dependencies
#   get_pkgs_currently_installed
#   get_pkgs_that_need_to_be_upgraded
#   get_pkgs_to_be_installed
#   get_pkgs_to_be_reinstalled
#   get_pkgs_to_be_removed
#   get_pkgs_to_be_upgraded
#   get_relative_softlink_target
#   get_rpm_pkgs_currently_installed
#   group_to_gid
#   _include_bundle
#   install_file
#   install_pkgs
#   md5sum_match
#   read_definition_file
#   reinstall_pkgs
#   remove_file
#   remove_pkgs
#   report_improper_file_definition
#   report_improper_service_definition
#   run_cmd
#   set_ownership_and_permissions
#   sync_state
#   turn_groupnames_into_gids
#   turn_service_into_file_entry
#   turn_usernames_into_uids
#   uid_gid_and_mode_match
#   _upgrade_ssm
#   user_is_root
#   user_to_uid
#   version
#
################################################################################


#
# Hash for holding package information
my %PKGS_FROM_STATE_DEFINITION;
my %PKGS_PROVIDED_BY_PKGS_FROM_STATE_DEFINITION;

#
# Hashes for holding file and service information
my (
    %TYPE,    # regular, block, character, fifo, softlink, or hardlink
              # this is the only hash where _every_ file will have an 
              # entry.  Use it to get a list of filenames.
    %MODE,
    %OWNER,
    %GROUP,
    %MD5SUM,
    %MAJOR,
    %MINOR,
    %TARGET,      # target file or directory for a link
    %PRESCRIPT,   # script or command to be run before installing a file
    %POSTSCRIPT,  # script or command to be run after installing a file
    %DEPENDS,     # package and, or file dependencies
    %DETAILS,     # runlevel information for services
    %PRIORITY,    # priority level for files
);

my $ERROR_LEVEL = 0;

################################################################################
#
#   Subroutines
#
sub read_definition_file {

    unless( defined($main::o{definition_file}) ) {
        
        my $file = '/etc/ssm/client.conf';
        open(FILE,"<$file") or die("Couldn't open $file for reading. $!\n");
            while(<FILE>) {
                if(m/^definition_file\s+(.*)(\s|#|$)/) {
                    $main::o{definition_file} = $1;
                }
            }
        close(FILE);
    }

    if( $main::o{definition_file} =~ m,/$, ) {
        # URI ends with a slash.  Is a dir.  Append hostname -f.
        my $hostname = `hostname -f`;
        chomp $hostname;

        $main::o{definition_file} .= $hostname;
    }

    print "\nState definition file:  $main::o{definition_file}\n\n" unless( $main::o{quiet} );

    my $tmp_file = get_file($main::o{definition_file});

    open(FILE,"<$tmp_file") or die "Couldn't open $tmp_file for reading: $!";
        my @input = (<FILE>);
    close(FILE);
    unlink $tmp_file;
    
    my $stanza_terminator = '^(\s+|$)';
    my $comment = '^#';

    #
    # Add a blank line at the end, so that we don't get an error when 
    # matching for a $stanza_terminator against an undef value.
    # Otherwise, it's harmless. -BEF-
    push(@input, "");   
    while(@input) {
        $_ = shift @input;

        # 
        # [global] section
        #
        if( m/^\[global\]/ ) {

            $_ = shift @input;
            until( m/$stanza_terminator/ ) {

                #
                # Allow "key = value" or "key=value" type definitions.
                s/\s*=\s*/ /o;

                if( m/^pkg_manager\s+(.*)(\s|#|$)/ )    { $main::o{pkg_manager} = $1; }
                if( m/^base_uri\s+(.*)(\s|#|$)/ )       { $main::o{base_uri} = $1; }
                if( m/^svn_uri\s+(.*)(\s|#|$)/ )        { $main::o{svn_uri} = $1; }

                $_ = shift @input;
            }

            if( ! defined $main::o{base_uri} ) { 
                # No base_uri defined.  Therefore, we assume base_uri 
                # should be the same as the definition file uri, sans
                # the filename itself. -BEF-
                $main::o{base_uri} = dirname( $main::o{definition_file} );
            }
        } 

        # 
        # [bundles] section
        #
        elsif( m/^\[bundles\]/ ) {

            $_ = shift @input;
            until( m/$stanza_terminator/ ) {

                if( m/$comment/ ) {
                    # do nothing
                }

                # Match only the first entry on the line.  This allows 
                # for comments after an entry. -BEF-
                elsif( m/^([\S]+)/ ) {
                    push( @input, _include_bundle($1) );
                }

                $_ = shift @input;
            }
        }

        # 
        # [packages] section
        #
        elsif( m/^\[packages\]/ ) {

            $_ = shift @input;
            until( m/$stanza_terminator/ ) {

                #
                # Allow "key = value" or "key=value" type definitions.
                s/\s*=\s*/ /o;

                if( m/$comment/ ) {
                    # do nothing
                }
                elsif( m/^[\S]+(\s+|=)/ ) {

                    my ($pkg, $ver) = split;
                    if( $ver eq '') { $ver = 'LATEST'; }
                    $PKGS_FROM_STATE_DEFINITION{$pkg} = $ver;
                }

                $_ = shift @input;
            }
        }

        # 
        # [service] sections
        #
        elsif( m/^\[service\]/ ) {

            my( $name, 
                $details,
                $depends,
                );

            $_ = shift @input;
            until( m/$stanza_terminator/ ) {
                
                #
                # Collect all the values
                chomp;

                #
                # Allow "key = value" or "key=value" type definitions.
                s/\s*=\s*/ /o;

                my ($key, $value) = split('\s+', $_, 2);

                if($key eq 'name') { 
                    $name = $value; 

                } elsif($key eq 'details') { 
                    # We don't remove spaces here, as they're part of the value
                    $details = $value;

                } elsif(($key eq 'depends') or ($key eq 'deps')) { 
                    # We don't remove spaces here, as they're part of the value
                    $depends = $value;

                }

                $_ = shift @input;
            }

            if( (defined $name) and (defined $DETAILS{$name}) ) {
                print STDERR "\n";
                print STDERR "ERROR: Multiple (conflicting) definitions for:\n";
                print STDERR "\n";
                print STDERR "  [service]\n";
                print STDERR "  name = $name\n";
                print STDERR "  ...\n";
                print STDERR "\n";
                print STDERR "  Exiting now without modifying the service. Please examine your state\n";
                print STDERR "  definition file and eliminate all but one of the definitions for\n";
                print STDERR "  this service.\n";
                print STDERR "\n";

                $ERROR_LEVEL++;

                # We go ahead and exit here to be super conservative.
                exit $ERROR_LEVEL;
            }

            #
            # Assign the values to hashes
            #
            if( defined $name ) {
                $DETAILS{$name} = $details if(defined $details);
                $DEPENDS{$name} = $depends if(defined $depends);
            }

            my $unsatisfied = check_depends($name);
            if($unsatisfied ne "1") {
                print "\n";
                print "Not OK: Service $name -> Unmet dependencies:\n";
                print "        $unsatisfied\n";
                print "\n";
                $ERROR_LEVEL++;

            } else {

                if(defined $DETAILS{$name}) {
                    turn_service_into_file_entry($name);
                } else {
                    report_improper_service_definition($name);
                }
            }
        }

        # 
        # [files] sections
        #
        elsif( m/^\[file\]/ ) {

            my( $name, 
                $type,
                $mode,
                $owner,
                $group,
                $md5sum,
                $major,
                $minor,
                $target,
                $prescript,
                $postscript,
                $depends,
                $priority,
                );

            $_ = shift @input;
            until( m/$stanza_terminator/ ) {
                
                #
                # Collect all the values
                chomp;

                #
                # Allow "key = value" or "key=value" type definitions.
                s/\s*=\s*/ /o;

                my ($key, $value) = split('\s+', $_, 2);

                if($key eq 'name') { 
                    $name = $value; 

                } elsif($key eq 'type')       { 
                    $value =~ s/\s.*//;
                    $type = $value;

                } elsif($key eq 'mode') { 
                    $value =~ s/\s.*//;
                    $mode = $value;

                } elsif($key eq 'owner') {
                    $value =~ s/\s.*//;
                    $owner = $value;

                } elsif($key eq 'group') {
                    $value =~ s/\s.*//;
                    $group = $value;

                } elsif($key eq 'md5sum') {
                    $value =~ s/\s.*//;
                    $md5sum = $value;

                } elsif($key eq 'major') {
                    $value =~ s/\s.*//;
                    $major = $value;

                } elsif($key eq 'minor') {
                    $value =~ s/\s.*//;
                    $minor = $value;

                } elsif($key eq 'target') {
                    $target = $value;

                } elsif($key eq 'prescript') {
                    $prescript = $value;

                } elsif($key eq 'postscript') { 
                    $postscript = $value;

                } elsif(($key eq 'depends') or ($key eq 'deps')) { 
                    $depends = $value;

                } elsif($key eq 'priority') { 
                    if( $value =~ /^\d+$/ ) {
                        $priority = $value;
                    } else {
                        # Easter egg or bug -- you decide...  You can use a value like "two" which is translated to "3". ;-)  heh, heh, heh... -BEF-
                        $priority = length $value;
                    }
                }

                $_ = shift @input;
            }

            # If no priority is set for this file, or the priority field is blank, use the default of zero.
            if(! defined $priority) {
                $priority = 0;
            }

            # If existing priority is higher than this file's priority
            if( (defined $PRIORITY{$name}) and ($PRIORITY{$name} > $priority) ) {
                # do nothing;

            # If existing priority is equal to this file's priority
            } elsif( (defined $PRIORITY{$name}) and ($PRIORITY{$name} == $priority) ) {
                # error out;
                print STDERR "\n";
                print STDERR "ERROR: Multiple (conflicting) definitions for:\n";
                print STDERR "\n";
                print STDERR "  [file]\n";
                print STDERR "  name     = $name\n";
                print STDERR "  priority = $priority\n";
                print STDERR "  ...\n";
                print STDERR "\n";
                print STDERR "  Exiting now without modifying the file. Please examine your state\n";
                print STDERR "  definition file and eliminate all but one of the definitions for\n";
                print STDERR "  this file, or raise the priority of one of the definitions.\n";
                print STDERR "\n";

                $ERROR_LEVEL++;

                # We go ahead and exit here to be super conservative.
                exit $ERROR_LEVEL;

            # If either:
            #   a) this file name has no existing priority
            #   b) this file's priority is higher than the existing priority
            } else {
                #
                # Assign the values to hashes
                #
                if( defined $name ) {
                    $TYPE{$name}       = $type       if(defined $type);
                    $MODE{$name}       = $mode       if(defined $mode);
                    $OWNER{$name}      = $owner      if(defined $owner);
                    $GROUP{$name}      = $group      if(defined $group);
                    $MD5SUM{$name}     = $md5sum     if(defined $md5sum);
                    $MAJOR{$name}      = $major      if(defined $major);
                    $MINOR{$name}      = $minor      if(defined $minor);
                    $TARGET{$name}     = $target     if(defined $target);
                    $PRESCRIPT{$name}  = $prescript  if(defined $prescript);
                    $POSTSCRIPT{$name} = $postscript if(defined $postscript);
                    $DEPENDS{$name}    = $depends    if(defined $depends);
                    $PRIORITY{$name}   = $priority   if(defined $priority);

                }

            }




            report_improper_file_definition($name) unless(defined $TYPE{$name});
        }
    }  

    print "\n" unless( $main::o{quiet} );

    turn_usernames_into_uids();
    turn_groupnames_into_gids();

    return 1;
}


sub turn_usernames_into_uids {
    foreach(keys %OWNER) {
        $OWNER{$_} = user_to_uid($OWNER{$_});
    }
    return 1;
}


sub turn_groupnames_into_gids {
    foreach(keys %GROUP) {
        $GROUP{$_} = group_to_gid($GROUP{$_});
    }
    return 1;
}


sub user_to_uid {

    my $user = shift;

    if($user =~ m/^\d+$/) {
        # it's already all-numeric; as in, a uid was specified in the definition
        return $user;
    } else {
        return (getpwnam $user)[2];
    }
}


sub group_to_gid {

    my $group = shift;

    if($group =~ m/^\d+$/) {
        # it's already all-numeric; as in, a gid was specified in the definition
        return $group;
    } else {
        return (getgrnam $group)[2];
    }
}


sub sync_state {

    _upgrade_ssm();

    #
    # Files
    foreach my $file (sort keys %TYPE) {

        my $unsatisfied = check_depends($file);
        if($unsatisfied ne "1") {
            print "\n";
            print "Not OK: File $file -> Unmet dependencies:\n";
            print "        $unsatisfied\n";
            print "\n";
            $ERROR_LEVEL++;
        }

        elsif( $TYPE{$file} eq 'softlink' ) {
            do_softlink($file);
        }
        elsif( $TYPE{$file} eq 'hardlink' ) {
            do_hardlink($file);
        }
        elsif(     $TYPE{$file} eq 'block' 
                or $TYPE{$file} eq 'character'
                or $TYPE{$file} eq 'fifo'      ) {
            do_special_file($file);
        }
        elsif( $TYPE{$file} eq 'regular' ) {
            do_regular_file($file);
        }
        elsif( $TYPE{$file} eq 'directory' ) {
            do_directory($file);
        }
        elsif( $TYPE{$file} eq 'unwanted' ) {
            do_unwanted_file($file);
        }
        else {
            report_improper_file_definition($file);
        }
    }
    

    # Get integer value that represents the number of packages specified.
    if( (scalar (keys %PKGS_FROM_STATE_DEFINITION)) == 0) {
        print "OK: Packages -> No [packages] specified in the state definition file.\n";
        return $ERROR_LEVEL;

    } else {

        # Do this here, and only once, for performance purposes. -BEF-
        %PKGS_PROVIDED_BY_PKGS_FROM_STATE_DEFINITION = get_pkgs_provided_by_pkgs_from_state_definition();

        if( $main::o{debug} ) {
            print '>> Contents of: %PKGS_PROVIDED_BY_PKGS_FROM_STATE_DEFINITION' . "\n";
            foreach (keys %PKGS_PROVIDED_BY_PKGS_FROM_STATE_DEFINITION) {
                print ">> Provided: $_\n" ;
            }
        }
    }

    ####################################################################
    #
    # Package related activities
    #
    #   Packages to remove:
    #   Remove dependencies (in definition):
    #       Shall I?
    #       do_it()
    #   
    #   Packages to Upgrade:
    #   Upgrade dependencies (not in definition):
    #       Shall I?
    #       do_it()
    #
    #   Packages to Install:
    #   Install dependencies (not in definition):
    #       Shall I?
    #       do_it()
    #
    #   Packages to Re-install:
    #       Shall I?
    #       do_it()
    #

    # 1 of 4
    my @pkgs_to_be_removed = get_pkgs_to_be_removed();
    my @pkgs_to_be_removed_deps;

    my $do_remove = undef;
    if($main::o{yes}) {
        $do_remove = 1;
    } else {
        if( scalar(@pkgs_to_be_removed) > 0) {

            my %pkgs_currently_installed = get_pkgs_currently_installed();
            my %reverse_dependencies = get_pkg_reverse_dependencies(@pkgs_to_be_removed);

            #
            # For each package that is not in the definition, and is about
            # to be removed, check to see if it is also listed in the reverse
            # dependencies.  If it is, delete it from the reverse dependencies,
            # as there's no need in having it listed in both places. -BEF-
            foreach my $pkg (@pkgs_to_be_removed) {
                if( defined($reverse_dependencies{$pkg}) ) {
                    delete $reverse_dependencies{$pkg};
                }
            }

            #
            # For each of the reverse dependencies, if the package is 
            # currently installed, include it as a package to remove.
            # We don't include other dependencies, as there's no reason
            # to try and remove something that isn't there. -BEF-
            foreach my $pkg (keys %reverse_dependencies) {
                if( defined($pkgs_currently_installed{$pkg}) ) {
                    push(@pkgs_to_be_removed_deps, $pkg);
                }
            }

            print "\n";

            print "Packages to remove:\n";
            print "------------------------------------------------------------------------\n";
            foreach(sort @pkgs_to_be_removed) {
                print "$_\n";
            }
            print "\n";

            print "Packages to remove (that depend on the above):\n";
            print "------------------------------------------------------------------------\n";
            foreach(sort @pkgs_to_be_removed_deps) {
                print "$_\n";
            }
            print "\n";

            print "  Shall I remove these packages? [N/y]: ";
            if( do_you_want_me_to() eq 'yes' ) { 
                $do_remove = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }
        } else {
            print "OK: Packages -> No packages to remove.";
        }
        print "\n";
    }

    if( defined($do_remove) ) {
        remove_pkgs(@pkgs_to_be_removed, @pkgs_to_be_removed_deps);
    }



    # 2 of 4
    my @pkgs_to_be_upgraded = get_pkgs_to_be_upgraded();
    my @pkgs_to_be_upgraded_deps;

    my $do_upgrade = undef;
    if($main::o{yes}) {
        $do_upgrade = 1;
    } else {
        if( scalar(@pkgs_to_be_upgraded) > 0) {

            my %dependencies = get_pkg_dependencies(@pkgs_to_be_upgraded);
            foreach $_ (keys %dependencies) {
                my $already_listed = 'no';

                #
                # Each package may have a series of alternates as dependencies,
                # any one of which will satisfy it's dependency.  Those are
                # stored as the value in %dependencies separated by spaces. -BEF-
                #
                my @pkg_alternatives = split(/\s+/, $_);
                foreach my $pkg (@pkg_alternatives) {
                    if((defined $PKGS_FROM_STATE_DEFINITION{$pkg}) or (defined $PKGS_PROVIDED_BY_PKGS_FROM_STATE_DEFINITION{$pkg})) {
                        $already_listed = 'yes';
                    }
                }
                if($already_listed eq 'no') {
                    #
                    # If we find that a package has a dependency that is not in it's state
                    # definition, then we add it to the list of dependencies of packages to
                    # be upgraded.  If there are alternates, we just take the first one. -BEF-
                    push(@pkgs_to_be_upgraded_deps, $pkg_alternatives[0]);
                }
            }

            print "\n";

            print "Packages to upgrade:\n";
            print "------------------------------------------------------------------------\n";
            foreach(sort @pkgs_to_be_upgraded) {
                print "$_\n";
            }
            print "\n";

            print "Packages to upgrade (upgrade dependencies not in definition):\n";
            print "------------------------------------------------------------------------\n";
            foreach(sort @pkgs_to_be_upgraded_deps) {
                print "$_\n";
            }
            print "\n";

            print "  Shall I upgrade these packages? [N/y]: ";
            if( do_you_want_me_to() eq 'yes' ) { 
                $do_upgrade = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }

        } else {
            print "OK: Packages -> No packages to upgrade.";
        }
        print "\n";
    }

    if( defined($do_upgrade) ) {
        install_pkgs(@pkgs_to_be_upgraded, @pkgs_to_be_upgraded_deps);
    }


    # 3 of 4
    my @pkgs_to_be_installed = get_pkgs_to_be_installed();
    my @pkgs_to_be_installed_deps;

    my $do_install = undef;
    if($main::o{yes}) {
        $do_install = 1;
    } else {
        if( scalar(@pkgs_to_be_installed) > 0) {

            my %dependencies = get_pkg_dependencies(@pkgs_to_be_installed);
            foreach $_ (keys %dependencies) {
                my $already_listed = 'no';

                #
                # Each package may have a series of alternates as dependencies,
                # any one of which will satisfy it's dependency.  Those are
                # stored as the value in %dependencies separated by spaces. -BEF-
                #
                my @pkg_alternatives = split(/\s+/, $_);
                foreach my $pkg (@pkg_alternatives) {
                    if((defined $PKGS_FROM_STATE_DEFINITION{$pkg}) or (defined $PKGS_PROVIDED_BY_PKGS_FROM_STATE_DEFINITION{$pkg})) {
                        $already_listed = 'yes';
                    }
                }
                if($already_listed eq 'no') {
                    #
                    # If we find that a package has a dependency that is not in it's state
                    # definition, then we add it to the list of dependencies of packages to
                    # be upgraded.  If there are alternates, we just take the first one. -BEF-
                    push(@pkgs_to_be_installed_deps, $pkg_alternatives[0]);
                }
            }

            print "\n";

            print "Packages to install:\n";
            print "------------------------------------------------------------------------\n";
            foreach(sort @pkgs_to_be_installed) {
                print "$_\n";
            }
            print "\n";

            print "Packages to install (dependencies not in definition):\n";
            print "------------------------------------------------------------------------\n";
            foreach(sort @pkgs_to_be_installed_deps) {
                print "$_\n";
            }
            print "\n";

            print "  Shall I install these packages? [N/y]: ";
            if( do_you_want_me_to() eq 'yes' ) { 
                $do_install = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }

        } else {
            print "OK: Packages -> No packages to install.";
        }
        print "\n";
    }

    if( defined($do_install) ) {
        install_pkgs(@pkgs_to_be_installed, @pkgs_to_be_installed_deps);
    }


    # 4 of 4
    my @pkgs_to_be_reinstalled = get_pkgs_to_be_reinstalled();
    my $do_reinstall = undef;
    if($main::o{yes}) {
        $do_reinstall = 1;
    } else {
        if( scalar(@pkgs_to_be_reinstalled) > 0) {
            print "\n";
            print "Packages to re-install:\n";
            print "------------------------------------------------------------------------\n";
            foreach(@pkgs_to_be_reinstalled) {
                print "$_\n";
            }
            print "  Shall I re-install these packages? [N/y]: ";
            if( do_you_want_me_to() eq 'yes' ) { 
                $do_reinstall = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }
        } else {
            print "OK: Packages -> No packages to re-install.";
        }
        print "\n";
    }

    if( defined($do_reinstall) ) {
        reinstall_pkgs(@pkgs_to_be_reinstalled);
    }
    #
    ####################################################################

    # Remove checked out SSM DB, if it exists.
    my $ou_path = "/tmp/ssm_db.repo.$$";
    rmtree("$ou_path");

    return $ERROR_LEVEL;
}


sub get_pkg_reverse_dependencies {
    
    my @packages = @_;

    if( $main::o{pkg_manager} eq 'dpkg' ) {
        return get_deb_pkg_reverse_dependencies(@packages);
    }

    return undef;
}


sub get_pkg_dependencies {
    
    my @packages = @_;

    if( $main::o{pkg_manager} eq 'dpkg' ) {
        return get_deb_pkg_dependencies(@packages);
    }

    return undef;
}


sub get_pkgs_provided_by_pkgs_from_state_definition {

    my @packages = @_;

    if( $main::o{pkg_manager} eq 'dpkg' ) {
        return get_deb_pkg_provides(keys %PKGS_FROM_STATE_DEFINITION);
    }

    return undef;
}


sub get_deb_pkg_reverse_dependencies {

    my @packages = @_;

    my %reverse_dependencies;

    foreach my $pkg (@packages) {

        #
        # Find the reverse dependencies
        my $cmd = "apt-cache rdepends $pkg";
        print ">> $cmd\n" if( $main::o{debug} );
        my @rdeps;
        open(OUTPUT,"$cmd|") or die;
            while(<OUTPUT>) {
                chomp;
                #
                # Match this:
                #   package
                #
                # or this:
                #   |package
                #
                if(m/^\s+[|]?(\S+)/) {
                    push @rdeps, $1;
                }
            }
        close(OUTPUT);

        #
        # Is package really a dependency of each reverse dependency?  If
        # not, remove the reverse dependency.
        #
        # Now take the results for this particular package, and add them
        # to the compiled hash of results.
        foreach my $rdep (@rdeps) {

            #
            # get dependencies of reverse dependencies
            #
            # Unfortunately, rdepends includes Suggests: (and maybe 
            # Replaces:) entries as well as Depends:.  So now we must 
            # query the resulting reverse-dependencies for their 
            # dependencies to see if $pkg is actually listed there.  If 
            # not, then it must just be a Recommends: or Suggests:.
            my %dependencies = get_deb_pkg_dependencies($rdep);

            my @pkg_alternatives;
            foreach $_ (keys %dependencies) {
                push @pkg_alternatives, split(/\s+/, $_);
            }
            foreach my $dependency (@pkg_alternatives) {
                if($dependency eq $pkg) {
                    $reverse_dependencies{$rdep} = 1; 
                }
            }
        }
    }

    return %reverse_dependencies;
}


sub get_deb_pkg_provides {

    my @packages = @_;

    my $cmd = 'apt-cache show';
    foreach my $pkg (@packages) {
        $cmd .= " $pkg"
    }

    print ">> $cmd\n" if( $main::o{debug} );
    my @array;
    open(OUTPUT,"$cmd|") or die;
        while(<OUTPUT>) {
            if(s/^Provides://) {
                push @array, split(/,/);
            }
        }
    close(OUTPUT);

    my %provides;
    foreach(@array) {
        chomp;
        s/^\s+//;
        s/\s.*//;   
        $provides{$_} = 1;
    }

    return %provides;
}


#
# Returns a hash of packages, with the package names as the keys.
#
# NOTE:  If there are alternates as dependencies, all alternates are returned
#        as the a key in a space seperated string format. -BEF-
#
sub get_deb_pkg_dependencies {

    my @packages = @_;

    my $cmd = 'apt-cache show';
    foreach my $pkg (@packages) {
        $cmd .= " $pkg"
    }

    print ">> $cmd\n" if( $main::o{debug} );
    my @array;
    open(OUTPUT,"$cmd|") or die;
        while(<OUTPUT>) {
            if(s/^Depends://) {
                push @array, split(/,/);
            }
        }
    close(OUTPUT);

    my %dependencies;
    foreach(@array) {
        chomp;
        s/^\s+//;
        
        ## If there's an alternate dependency "this | that", this only 
        ## gets the first one.  May want to make this whole function
        ## more comprehensive at some point. -BEF-
        #s/\s.*//;   

        # If there's an alternate dependency "this | that", this
        # get's all of them as a space separated list. -BEF-
        s/ \(= \S+\)( \|)?//g;
        $dependencies{$_} = 1;
    }

    return %dependencies;
}


sub get_pkgs_to_be_removed {

    my %pkgs_currently_installed = get_pkgs_currently_installed();

    my @array;
    foreach my $pkg ( keys %pkgs_currently_installed ) {
        if( ! defined($PKGS_FROM_STATE_DEFINITION{$pkg}) ) {
            push @array, $pkg;
        }
    }
    
    return @array;
}


#
# returns an array:  "pkg=version" or just "pkg"
#   - packages not currently installed
#   - packages from definition include version numbers if appropriate
#
sub get_pkgs_to_be_installed {

    my %pkgs_currently_installed = get_pkgs_currently_installed();

    #
    # If it's in the state definition, but not installed, install it.
    #
    my %hash;
    foreach my $pkg (keys %PKGS_FROM_STATE_DEFINITION) {
        if( $PKGS_FROM_STATE_DEFINITION{$pkg} and ! $pkgs_currently_installed{$pkg} ) {
            $hash{$pkg} = $PKGS_FROM_STATE_DEFINITION{$pkg};
        }
    }

    my @array;
    foreach my $pkg (keys %hash) {
        
        #
        # If this package includes version information, use it.  
        # Otherwise, we'll just use the package name.
        #

        if( $PKGS_FROM_STATE_DEFINITION{$pkg} ne 'LATEST' ) {
            $_ = "$pkg=$PKGS_FROM_STATE_DEFINITION{$pkg}";
        } else {
            $_ = $pkg;
        }

        push( @array, $_ );
    }

    return @array;
}


#
# returns an array:  "pkg=version" or just "pkg"
#   - packages currently installed that need to be upgraded
#   - upgrade dependencies
#
sub get_pkgs_to_be_upgraded {

    my %hash = get_pkgs_that_need_to_be_upgraded();
    
    my @array;
    foreach my $pkg (keys %hash) {
    
        # If a specific version is specified by the definition,
        #  - include entry with version

        if( (defined $PKGS_FROM_STATE_DEFINITION{$pkg}) and ($PKGS_FROM_STATE_DEFINITION{$pkg} ne 'LATEST') ) {
            $_ = "$pkg=$PKGS_FROM_STATE_DEFINITION{$pkg}";
        } else {
            $_ = $pkg;
        }
    
        push( @array, $_ );
    }
    
    return @array;
}


sub get_pkgs_to_be_reinstalled {

    #
    # returns an array:  "pkg=version" or just "pkg"
    #   packages currently installed that need to be reinstalled
    #

    my @array;

    #
    # If a package exists in the state definition:
    #   - and is versioned, but a different version is already installed, re-install it
    # 
    my %pkgs_currently_installed = get_pkgs_currently_installed();
    foreach my $pkg (keys %pkgs_currently_installed) {
        #
        # It's possible at this point, that a package on the system is
        # not in the state definition.  So, our first test below needs
        # to check for the existence of the package in the state
        # definition, so that we don't assume and produce error
        # messages from the code.
        #
        if( defined $PKGS_FROM_STATE_DEFINITION{$pkg}
            and $PKGS_FROM_STATE_DEFINITION{$pkg} ne 'LATEST'
            and $PKGS_FROM_STATE_DEFINITION{$pkg} ne $pkgs_currently_installed{$pkg}
            ) {
            push( @array, "$pkg=$PKGS_FROM_STATE_DEFINITION{$pkg}" );
        }
    }

    return @array;
}


sub get_pkgs_currently_installed {

    # hash: pkg = key; version = value

    if( ! defined($main::o{pkg_manager}) ) {
        print STDERR "\n";
        print STDERR "ERROR: Missing entry in definition file.  Please specify your package\n";
        print STDERR "       type with an entry like:\n";
        print STDERR "\n";
        print STDERR "       pkg_manager = dpkg\n";
        print STDERR "       or pkg_manager = up2date\n";
#        print STDERR "       or pkg_manager = yum\n";
#        print STDERR "       or pkg_manager = apt-rpm\n";
        print STDERR "\n";

        exit 1;
    }

    if( $main::o{pkg_manager} eq 'dpkg' ) {
        return get_deb_pkgs_currently_installed();
    }
    elsif( $main::o{pkg_manager} eq 'up2date' ) {
        return get_rpm_pkgs_currently_installed();
    }

    return undef;
}


sub get_pkgs_that_need_to_be_upgraded {

    # hash: pkg = key; version = value

    my %hash;
    if($main::o{pkg_manager} eq 'dpkg') {
        %hash = get_deb_pkgs_that_need_to_be_upgraded();
    }

    return %hash;
}


sub remove_pkgs {

    my @pkgs = @_;

    my $cmd;
    if( $main::o{pkg_manager} eq 'dpkg' ) {
        foreach(@pkgs) {
            next unless(defined $_);
            $cmd .= " $_";
        }
    }

    if(defined $cmd) {
        print "FIXING: Packages -> Removing.\n" unless($main::o{quiet});
        $cmd = 'DEBIAN_FRONTEND=noninteractive apt-get -q=2 --force-yes -y remove' . $cmd;
        run_cmd($cmd);
    }

    return 1;
}


sub install_pkgs {

    my @pkgs = @_;

    if($main::o{pkg_manager} eq 'dpkg') {

        my $pkgs;
        foreach(@pkgs) {
            next unless(defined $_);
            $pkgs .= " $_";
        }

        if(defined $pkgs) {

            my $cmd;

            print "FIXING: Packages -> Downloading.\n" unless($main::o{quiet});
            $cmd = 'DEBIAN_FRONTEND=noninteractive apt-get -q=2 --force-yes -y --download-only install';
            run_cmd($cmd);

            print "FIXING: Packages -> Installing.\n" unless($main::o{quiet});
            $cmd = 'DEBIAN_FRONTEND=noninteractive apt-get -q=2 --force-yes -y install' . $pkgs;
            run_cmd($cmd);

            # Remove any packages lying around in the cache.  Again.
            $cmd = 'apt-get clean';
            run_cmd($cmd);
        }
    }

    return 1;    
}


sub reinstall_pkgs {

    my @pkgs = @_;

    if($main::o{pkg_manager} eq 'dpkg') {

        my $cmd;
        foreach(@pkgs) {
            next unless(defined $_);
            $cmd .= " $_";
        }

        if(defined $cmd) {
            print "Downloading packages...\n" unless($main::o{quiet});
            $cmd = 'DEBIAN_FRONTEND=noninteractive apt-get -q=2 --force-yes -y --reinstall --download-only install' . $cmd;
            run_cmd($cmd);

            print "Re-installing packages...\n" unless($main::o{quiet});
            $cmd = 'DEBIAN_FRONTEND=noninteractive apt-get -q=2 --force-yes -y --reinstall install' . $cmd;
            run_cmd($cmd);

            # Remove any packages lying around in the cache.  Again.
            $cmd = 'apt-get clean';
            run_cmd($cmd);
        }
    }

    return 1;    
}


sub check_in {

    return 0;
}


sub version {
    my $progname = basename($0);
    my $VERSION = '___VERSION___';
    print <<EOF;
$progname (part of System State Manager) v$VERSION
    
EOF
}


#
# Usage: my $arch = get_arch();
#
sub _get_arch {

        use POSIX;

	my $arch = (uname())[4];
	$arch =~ s/i.86/i386/;

	return $arch;
}


#
# Usage:  
#       run_cmd("my shell command");
#       run_cmd("my shell command", 1 );
#       run_cmd("my shell command", 1, 1 );
#       run_cmd("my shell command", undef, 1);
#       run_cmd("my shell command", , 1);
#       run_cmd("my shell command", , );
#
#       Do not use a "zero" (0) to indicate off, use a blank or "undef" (undef).
#
#       First argument:  the "command" to run.
#           Required.
#
#       Second argument:  '1' to print a newline after the command.
#           Defaults to "undef".
#
#       Third argument:  '1' to do this command _even_ if --dry-run is specified.
#           Defaults to "undef".
#
sub run_cmd {

        my $cmd             = shift;
        my $add_newline     = shift;
        my $even_if_dry_run = shift;

        print ">> $cmd\n" if( $main::o{debug} );
        if( ! $main::o{dry_run} or defined($even_if_dry_run) ) { 
            !system($cmd) or die("FAILED: $cmd\n $!");
        }

        unless( $main::o{quiet} ) { print "\n" if( defined $add_newline ); }

        return 1;
}


sub get_deb_pkgs_that_need_to_be_upgraded {

    my %hash;
    my $cmd;

    #
    # Get the latest updates
    print "OK: Packages -> Updating availability information.\n" unless($main::o{quiet});
    $cmd = 'apt-get -q=2 update';
    run_cmd($cmd);

    #
    # Get a list of packages that would be upgraded
    $cmd = 'DEBIAN_FRONTEND=noninteractive apt-get -s dist-upgrade';
    print ">> $cmd\n" if( $main::o{debug} );

    #
    # This is harmless, so we run it even if --dry-run, so that we get
    # the output that is interesting for the rest of the dry run.
    #
    open(OUTPUT,"$cmd|");
        while(<OUTPUT>) {
            if( m/^Inst\s(\S+)\s+/ ) {
                $hash{$1} = 'LATEST';
            }
        }
    close(OUTPUT);

    return %hash;

}


sub get_rpm_pkgs_currently_installed {

    #
    # returns a hash: package => version
    #

    my %hash;

    # 
    # rpm --queryformat reference:
    #   http://docs.fedoraproject.org/drafts/rpm-guide-en/ch04s02.html#id2941528
    #
    #XXX consider: -%{RELEASE}
    my $cmd = 'rpm -qa --queryformat %{NAME}\ \ %{VERSION}\\\n';
    print ">> $cmd\n" if( $main::o{debug} );
    open(FILE,"$cmd|") or die("couldn't open $cmd for reading");
    while (<FILE>) {

            #
            # Only choose packages marked as installed (ii)
            #
            # Sample output:
            #
            #   ii  abcde         2.3.99.2-1         A Better CD Encoder
            #   ii  acpi          0.09-1             displays information on ACPI devices
            #   ii  acpi-support  0.73               a collection of useful events for acpi
            #   ii  acpid         1.0.4-1ubuntu10    Utilities for using ACPI power management
            #   ii  acroread      7.0.1-0.0.ubuntu1  Adobe Acrobat Reader: Portable Document Form
            #   ii  adduser       3.80ubuntu2        Add and remove users and groups
            #
            #                   Matches the package name -> $1
            #                   | 
            #                   |       Matches the version string -> $2
            #                   |       |
            #                   vvvvv   vvvvv
            next unless(m/^(\S+)\s+(\S+)/);

            $hash{$1} = $2;
    }
    close(FILE);

    return %hash;
}


sub get_deb_pkgs_currently_installed {

    #
    # returns a hash: package => version
    #

    my %hash;

    my $cmd = 'dpkg -l';
    print ">> $cmd\n" if( $main::o{debug} );
    open(FILE,"$cmd|") or die("couldn't open $cmd for reading");
    while (<FILE>) {

            #
            # Only choose packages marked as installed (ii)
            #
            # Sample output:
            #
            #   ii  abcde         2.3.99.2-1         A Better CD Encoder
            #   ii  acpi          0.09-1             displays information on ACPI devices
            #   ii  acpi-support  0.73               a collection of useful events for acpi
            #   ii  acpid         1.0.4-1ubuntu10    Utilities for using ACPI power management
            #   ii  acroread      7.0.1-0.0.ubuntu1  Adobe Acrobat Reader: Portable Document Form
            #   ii  adduser       3.80ubuntu2        Add and remove users and groups
            #
            #                   Matches the package name -> $1
            #                   | 
            #                   |       Matches the version string -> $2
            #                   |       |
            #                   vvvvv   vvvvv
            next unless(m/^ii\s+(\S+)\s+(\S+)\s+/);

            $hash{$1} = $2;
    }
    close(FILE);

    return %hash;
}


sub do_you_want_me_to {

    my $msg = shift;

    print $msg if(defined $msg);

    $_ = <STDIN>;

    if( m/^(y|yes)/i ) {
        return 'yes';
    } elsif( m/^(d|diff)/i ) {
        return 'diff';
    } elsif( m/^(a|add)/i ) {
        return 'add';
    } elsif( m/^(n|no)/i ) {
        return 'no';
    }
}


sub do_softlink {

    my $file = shift;

    #
    # validate input
    unless( 
            ($file            =~ m#^/#)
        and ($TARGET{$file} =~ m#^/#)
        and ($TYPE{$file}   =~ m#\S#)
    ) {
        report_improper_file_definition($file);
    }

    #
    # Handle double slashes in target names
    $TARGET{$file} =~ s#/+#/#g;

    my $relative_target = get_relative_softlink_target($file, $TARGET{$file});
    my $current_target = readlink($file);

    #
    # Possibilities are:
    #   - $file exists and $current_target is right
    #       - leave it alone
    #
    #   - $file doesn't exist
    #       - create link
    #
    #   - $file exists, $current_target is wrong
    #       - rm $file 
    #       - create link
    #
    #   - $file exists, $current_target is undef (Ie: $file is not a softlink)
    #       - rm $file 
    #       - create link
    #
    unless( (defined $current_target) and ($current_target eq $relative_target) ) {

        my $fix_it = undef;

        if($main::o{yes}) {
            $fix_it = 1;

        } elsif( ! $main::o{quiet} ) {
            #
            # present user with options.
            print "\n";
            print "Not OK: Soft link $file -> $relative_target ($TARGET{$file})\n";
            print "  Need to:\n";
            print "  - $PRESCRIPT{$file}\n" if($PRESCRIPT{$file});
            print "  - remove $file\n" if( -e $file );
            print "  - create link\n";
            print "  - $POSTSCRIPT{$file}\n" if($POSTSCRIPT{$file});
            print "  Shall I do this now? [N/y]: ";

            if( do_you_want_me_to() eq 'yes' ) { 
                $fix_it = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }

        } else {
            $ERROR_LEVEL++;
        }

        if( defined($fix_it) and ! $main::o{dry_run} ) {

            print "FIXING: Soft link $file -> $relative_target ($TARGET{$file})\n" unless($main::o{quiet});

            if($PRESCRIPT{$file}) {
                my $cmd = $PRESCRIPT{$file};
                print qq(RUNNING: $cmd\n);
                run_cmd($cmd);
            }

            rmtree($file);
            symlink($relative_target, $file) or die "Couldn't symlink($relative_target, $file) $!";

            if($POSTSCRIPT{$file}) {
                my $cmd = $POSTSCRIPT{$file};
                print qq(RUNNING: $cmd\n);
                run_cmd($cmd);
            }

        }
    } else {
        print "OK: Soft link $file -> $relative_target ($TARGET{$file})\n" unless($main::o{quiet});
    }

    return 1;
}


sub get_relative_softlink_target {

    my $file = shift;
    my $target = shift;

    #
    # Get relative target
    my @file = split(/\//, $file);
    shift @file;
    
    my @target = split(/\//, $target);
    shift @target;
    
    my $i = 0;

    while( $file[$i] eq $target[$i] ) {
        $target[$i] = undef;
        $i++;
    }

    my $relative_target;
    foreach(@target) {
        if(defined $_) {
            $relative_target .= '/' . $_;
        }
    }

    if( $relative_target =~ m|/.*/| ) {
        # Our relative target has more than one slash; must be including 
        # at leaset one directory component -- add double dots.  Makes
        # for a link that looks like this:
        #
        #   my_link -> ../usr/lib/relative_target
        #
        $relative_target = '..' . $relative_target;

    } else {
        # Our relative target has only one slash; must be a single file;
        # strip slash.  Makes for a link like:
        #
        #   my_link -> relative_target
        #
        $relative_target =~ s|^/||;
    }

    return $relative_target;
}


sub do_special_file {

    my $file   = shift;

    #
    # validate input
    if( 
           ($file !~ m/\S/)
        or ($TYPE{$file} !~ m/\S/)
        or ($MODE{$file} !~ m/^\d+$/)
        or ($OWNER{$file} !~ m/\S/)
        or ($GROUP{$file} !~ m/\S/)
        or (
                (($TYPE{$file} eq 'character') or ($TYPE{$file} eq 'block')) 
                and 
                (($MAJOR{$file} !~ m/^\d+$/) or ($MINOR{$file} !~ m/^\d+$/))
           )
    ) {
        report_improper_file_definition($file);
    }

    #
    # Does it need fixing?
    my $needs_fixing = undef;
    if( ! -e $file ) {
        # Ain't there
        $needs_fixing = 1;
    } 
    else {

        my $st = stat($file);

        if( ! uid_gid_and_mode_match($file) ) {
            $needs_fixing = 1;
        } 
        elsif( ($TYPE{$file} eq 'fifo') and (! S_ISFIFO($st_mode)) ) {
            $needs_fixing = 1;
        }
        elsif( ($TYPE{$file} eq 'block') and (! S_ISBLK($st_mode)) ) {
            $needs_fixing = 1;
        }
        elsif( ($TYPE{$file} eq 'character') and (! S_ISCHR($st_mode)) ) {
            $needs_fixing = 1;
        }
    }

    #
    # Should we actually fix it?
    my $fix_it = undef;
    if( defined($needs_fixing) ) {
        if($main::o{yes}) {
            $fix_it = 1;
        } elsif( ! $main::o{quiet} ) {

            print "\n";
            print "Not OK: " . ucfirst($TYPE{$file}) . " file $file\n";
            print "  Need to:\n";
            print "  - $PRESCRIPT{$file}\n" if($PRESCRIPT{$file});
            print "  - remove $file\n" if( -e $file );
            print "  - create $file\n";
            print "  - $POSTSCRIPT{$file}\n" if($POSTSCRIPT{$file});
            print "  Shall I do this now? [N/y]: ";

            if( do_you_want_me_to() eq 'yes' ) { 
                $fix_it = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }

        } else {
            $ERROR_LEVEL++;
        }

    } else {
        print "OK: " . ucfirst($TYPE{$file}) . " file $file\n" unless($main::o{quiet});
    }

    #
    # Take action
    if( defined($fix_it) and ! $main::o{dry_run} ) {

        print "FIXING: " . ucfirst($TYPE{$file}) . " file $file\n" unless($main::o{quiet});

        if($PRESCRIPT{$file}) {
            my $cmd = $PRESCRIPT{$file};
            print qq(RUNNING: $cmd\n);
            run_cmd($cmd);
        }

        rmtree($file);

        if($TYPE{$file} eq 'fifo') {
            umask 0000;
            mkfifo($file, oct($MODE{$file})) or die "Couldn't mkfifo($file, $MODE{$file})$!";
        }
        elsif($TYPE{$file} eq 'character') {
            #
            # Thanks to Jim Pirzyk, author of Unix::Mknod, for getting back to
            # me with a documentation fix that allows me to use his code here.
            # -BEF- 2006.05.08
            #
            my $mode = oct($MODE{$file});
            mknod( $file, S_IFCHR|$mode, makedev($MAJOR{$file}, $MINOR{$file}) );
        }
        elsif($TYPE{$file} eq 'block') {
            my $mode = oct($MODE{$file});
            mknod( $file, S_IFBLK|$mode, makedev($MAJOR{$file}, $MINOR{$file}) );
        }

        #
        # Ensure proper ownership
        set_ownership_and_permissions($file);

        if($POSTSCRIPT{$file}) {
            my $cmd = $POSTSCRIPT{$file};
            print qq(RUNNING: $cmd\n);
            run_cmd($cmd);
        }
    }

    return 1;
}


sub set_ownership_and_permissions {

    my $file = shift;

    print "FIXING: Ownership and Perms: $file\n" unless($main::o{quiet});

    chown $OWNER{$file}, $GROUP{$file}, $file;
    chmod oct($MODE{$file}), $file;

    return 1;
}


sub do_unwanted_file {

    my $file   = shift;

    #
    # validate input
    if(    ($file            !~ m#^/#)
        or ($TYPE{$file}   !~ m/\S/)
    ) {
        report_improper_file_definition($file);
    }

    #
    # Does it need fixing?
    my $needs_fixing = undef;
    if( -e $file ) {
        # It exists, and must be destroyed!!!
        $needs_fixing = 1;
    } 

    #
    # Should we actually fix it?
    my $fix_it = undef;
    unless(defined($needs_fixing)) {
        print "OK: Unwanted file $file doesn't exist\n" unless($main::o{quiet});
    } else {
        if($main::o{yes}) {
            $fix_it = 1;
        } elsif( ! $main::o{quiet} ) {

            print "\n";

            if( -d $file ) {
                print "Not OK: Unwanted directory exists: $file\n";
                print "  Need to:\n";
                print "  - $PRESCRIPT{$file}\n" if($PRESCRIPT{$file});
                print "  - remove the contents of $file\n";
                print "  - remove $file\n";
                print "  - $POSTSCRIPT{$file}\n" if($POSTSCRIPT{$file});
            } else {
                print "Not OK: Unwanted file exists: $file\n";
                print "  Need to:\n";
                print "  - $PRESCRIPT{$file}\n" if($PRESCRIPT{$file});
                print "  - remove $file\n";
                print "  - $POSTSCRIPT{$file}\n" if($POSTSCRIPT{$file});
            }
            print "  Shall I do this now? [N/y]: ";

            if( do_you_want_me_to() eq 'yes' ) { 
                $fix_it = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }

        } else {
            $ERROR_LEVEL++;
        }
    }

    #
    # Take action
    if( defined($fix_it) and ! $main::o{dry_run} ) {

        print "FIXING: Removing: $file\n" unless($main::o{quiet});

        if($PRESCRIPT{$file}) {
            my $cmd = $PRESCRIPT{$file};
            print qq(RUNNING: $cmd\n);
            run_cmd($cmd);
        }

        remove_file($file);

        if($POSTSCRIPT{$file}) {
            my $cmd = $POSTSCRIPT{$file};
            print qq(RUNNING: $cmd\n);
            run_cmd($cmd);
        }
    }

    return 1;
}


sub do_directory {

    my $file   = shift;

    #
    # validate input
    if(    ($file            !~ m#^/#)
        or ($TYPE{$file}   !~ m/\S/)
        or ($MODE{$file}   !~ m/^\d+$/)
        or ($OWNER{$file}  !~ m/\S/)
        or ($GROUP{$file}  !~ m/\S/)
    ) {
        report_improper_file_definition($file);
    }

    #
    # Does it need fixing?
    my $needs_fixing = undef;
    my $just_fix_uid_gid_and_mode = undef;
    if( ! -e $file ) {
        # Ain't there
        $needs_fixing = 1;
    } 
    elsif( ! -d $file ) {
        # It's not a directory
        $needs_fixing = 1;
    } 
    elsif( ! uid_gid_and_mode_match($file) ) {
        $needs_fixing = 1;
        $just_fix_uid_gid_and_mode = 1;
    } 

    #
    # Should we actually fix it?
    my $fix_it = undef;
    if( defined($needs_fixing) ) {

        if($main::o{yes}) {
            $fix_it = 1;

        } elsif( ! $main::o{quiet} ) {

            print "\n";
            print "Not OK: Directory $file\n";
            print "  Need to:\n";
            if( defined($just_fix_uid_gid_and_mode) ) {
                print "  - fix ownership and permissions\n";
            } else {
                print "  - $PRESCRIPT{$file}\n" if($PRESCRIPT{$file});
                print "  - create $file\n";
                print "  - $POSTSCRIPT{$file}\n" if($POSTSCRIPT{$file});
            }

            my $msg = "  Shall I do this now? [N/y]: ";
            my $answer = do_you_want_me_to($msg);

            if( $answer eq 'yes' ) { 
                $fix_it = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }

        } else {
            $ERROR_LEVEL++;
        }

    } else {
        print "OK: Directory $file\n" unless($main::o{quiet});
    }

    #
    # Take action
    if( defined($fix_it) and ! $main::o{dry_run} ) {
        if( defined($just_fix_uid_gid_and_mode) ) {
            set_ownership_and_permissions($file);
        } else {
            create_directory($file);
        }
    }

    return 1;
}


sub do_regular_file {

    my $file   = shift;

    #
    # validate input
    if(    ($file            !~ m#^/#)
        or ($TYPE{$file}   !~ m/\S/)
        or ($MODE{$file}   !~ m/^\d+$/)
        or ($OWNER{$file}  !~ m/\S/)
        or ($GROUP{$file}  !~ m/\S/)
        or ($MD5SUM{$file} !~ m/\S/)
    ) {
        report_improper_file_definition($file);
    }

    #
    # Does it need fixing?
    my $needs_fixing = undef;
    my $just_fix_uid_gid_and_mode = undef;
    if( ! -e $file ) {
        # Ain't there
        $needs_fixing = 1;
    } 
    elsif( ! -f $file ) {
        # It's not a regular file
        $needs_fixing = 1;
    } 
    elsif( ! md5sum_match($file) ) {
        $needs_fixing = 1;
    }
    elsif( ! uid_gid_and_mode_match($file) ) {
        $needs_fixing = 1;
        $just_fix_uid_gid_and_mode = 1;
    } 

    #
    # Should we actually fix it?
    my $fix_it = undef;
    if( defined($needs_fixing) ) {

        if($main::o{yes}) {
            $fix_it = 1;

        } elsif( ! $main::o{quiet} ) {

            print "\n";
            print "Not OK: Regular file $file\n";
            print "  Need to:\n";
            if( defined($just_fix_uid_gid_and_mode) ) {
                print "  - fix ownership and permissions\n";
            } else {
                print "  - $PRESCRIPT{$file}\n" if($PRESCRIPT{$file});
                print "  - install $file\n";
                print "  - $POSTSCRIPT{$file}\n" if($POSTSCRIPT{$file});
            }

            my $msg = "  Shall I do this now? [N/y/d/a]: ";
            my $answer = do_you_want_me_to($msg);
            while( $answer eq 'diff' ) {
                diff_file($file);
                $answer = do_you_want_me_to($msg);
            }

            if( $answer eq 'yes' ) { 
                $fix_it = 1;
            } elsif( $answer eq 'add' ) { 
                _add_file( $file );
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }

        } else {
            $ERROR_LEVEL++;
        }

    } else {
        print "OK: Regular file $file\n" unless($main::o{quiet});
    }

    #
    # Take action
    if( defined($fix_it) and ! $main::o{dry_run} ) {
        if( defined($just_fix_uid_gid_and_mode) ) {
            set_ownership_and_permissions($file);
        } else {
            install_file($file);
        }
    }

    return 1;
}


sub md5sum_match {

    my $file = shift;

    open(FILE, "<$file") or die "Can’t open ’$file’ for reading: $!";
        binmode(FILE);
        if( Digest::MD5->new->addfile(*FILE)->hexdigest eq $MD5SUM{$file} ) {
            return 1;
        }
    close(FILE);

    return undef;
}


sub remove_file {

    my $file = shift;

    rmtree($file);

    return 1;
}


sub diff_file {

    my $file = shift;

    my $uri = qq($main::o{base_uri}/$file/$MD5SUM{$file});
    
    my $tmp_file = get_file($uri);

    my $cmd = "diff -urN $file $tmp_file | less";
    run_cmd($cmd, , 1);

    return 1;
}


sub create_directory {

    my $file = shift;

    print "FIXING: Creating: $file\n" unless($main::o{quiet});

    if($PRESCRIPT{$file}) {
        my $cmd = $PRESCRIPT{$file};
        print qq(RUNNING: $cmd\n);
        run_cmd($cmd);
    }

    if(-e $file) { unlink($file); }

    eval { mkpath($file) };
    if($@) {
        die "Couldn’t create $file: $@";
    }

    set_ownership_and_permissions($file);

    if($POSTSCRIPT{$file}) {
        my $cmd = $POSTSCRIPT{$file};
        print qq(RUNNING: $cmd\n);
        run_cmd($cmd);
    }

    return 1;
}


sub install_file {

    my $file = shift;

    print "FIXING: Installing: $file\n" unless($main::o{quiet});

    _backup($file);

    if($PRESCRIPT{$file}) {
        my $cmd = $PRESCRIPT{$file};
        print qq(RUNNING: $cmd\n);
        run_cmd($cmd);
    }

    my $uri = qq($main::o{base_uri}/$file/$MD5SUM{$file});
    
    my $tmp_file = get_file($uri);

    #XXX do we want to do rmtree?  Think $file=directory ; 
    #rmtree($file);
    copy($tmp_file, $file) or die "Failed to copy($tmp_file, $file): $!";
    unlink $tmp_file;

    set_ownership_and_permissions($file);

    if($POSTSCRIPT{$file}) {
        my $cmd = $POSTSCRIPT{$file};
        print qq(RUNNING: $cmd\n);
        run_cmd($cmd);
    }

    return 1;
}


#
# my $tmp_file = get_file($file);
#
sub get_file {

    # copies $file, from wherever, to a temporary file name
    # returns that temporary file name

    my $file = shift;

    my $tmp_file = choose_tmp_file();

    # remove multiple slashes anywhere but after a protocol specifier
    $file =~ s#([^:/])/+#$1/#g;

    if( ($file =~ m#^file://#) or ($file =~ m#^/#) ) {

        $file =~ s#file://#/#;
        $file =~ s/(\s+|#).*//;
        copy($file, $tmp_file) or die "get_file(): Failed to copy($file, $tmp_file): $!";

    } elsif( ($file =~ m#^http://#) or ($file =~ m#^https://#) or ($file =~ m#^ftp://#) ) {

        my $cmd = "wget -q $file -O $tmp_file";
        run_cmd($cmd);

    } else {

        print STDERR "I don't know how to acquire a file using the specified protocol:\n";
        print STDERR "  $file\n";
        print STDERR "\n";
        print STDERR "  You may want to verify that you have a valid 'base_uri' specified\n";
        print STDERR "  in your definition file.\n";
        print STDERR "\n";

        exit 1;
    }

    return $tmp_file;
}


#
# my $tmp_file = choose_tmp_file();
#
sub choose_tmp_file {

    my $count = 0;
    my $file = "/tmp/system-state-manager_tmp_file";

    while( -e "$file.$count" ) {
        $count++;
    }
    $file = "$file.$count";

    umask 0077;
    open(FILE,">$file") or die "Couldn't open $file for writing $!";
        print FILE "I am a little tmp file created by System State Manager.\n";
    close(FILE);
    
    return $file;
}


sub do_hardlink {

    my $file   = shift;

    #
    # validate input
    unless( 
            ($file            =~ m#^/#)
        and ($TARGET{$file} =~ m#^/#)
        and ($TYPE{$file}   =~ m/\S/)
    ) {
        report_improper_file_definition($file);
    }

    #
    # Does it need fixing?
    my $needs_fixing = undef;
    if( ! -e $TARGET{$file} ) {
        # Target ain't there
        unless($main::o{quiet}) {
            print "\n";
            print "WARNING: Hard link not OK: $file\n";
            print "         Target doesn't exist:  $TARGET{$file}\n";
            print "         Auto-skipping.  Incrementing ERROR_LEVEL.\n";
            print "\n";
        }
        $ERROR_LEVEL++;
        return 1;
    } 
    elsif( ! -e $file ) {
        # Ain't there
        $needs_fixing = 1;
    } 
    elsif( -l $file ) {
        # It's a softlink
        $needs_fixing = 1;
    } 
    else {
        #
        # See if link's inode is the same as target's inode
        my $st;

        $st = stat($file);
        my $file_inode = $st_ino;

        $st = stat($TARGET{$file});
        my $target_inode = $st_ino;

        if($file_inode != $target_inode) {
            $needs_fixing = 1;
        }
    }

    #
    # Should we actually fix it?
    my $fix_it = undef;
    if( defined($needs_fixing) ) {

        if($main::o{yes}) {
            $fix_it = 1;

        } elsif( ! $main::o{quiet} ) {

            print "\n";
            print "Not OK: Hard link $file -> $TARGET{$file}\n";
            print "  Need to:\n";
            print "  - $PRESCRIPT{$file}\n" if($PRESCRIPT{$file});
            print "  - remove $file\n" if( -e $file );
            print "  - create link\n";
            print "  - $POSTSCRIPT{$file}\n" if($POSTSCRIPT{$file});
            print "  Shall I do this now? [N/y]: ";

            if( do_you_want_me_to() eq 'yes' ) { 
                $fix_it = 1;
            } else {
                print "  Ok, skipping this step.\n";
                sleep 1;
            }
        } else {
            $ERROR_LEVEL++;
        }

    } else {
        print "OK: Hard link $file -> $TARGET{$file}\n" unless($main::o{quiet});
    }

    #
    # Take action
    if( defined($fix_it) and ! $main::o{dry_run} ) {

        print "FIXING: Hard link $file -> $TARGET{$file}\n" unless($main::o{quiet});

        if($PRESCRIPT{$file}) {
            my $cmd = $PRESCRIPT{$file};
            print qq(RUNNING: $cmd\n);
            run_cmd($cmd);
        }

        rmtree($file);
        link($TARGET{$file}, $file) or die "Couldn't link($TARGET{$file}, $file) $!";
        #
        #XXX
        #   Should we accept and use owner, group, and mode info for hardlinks?
        #
        #   If perms and ownership are changed on a hardlink, they are 
        #   changed for the file itself, and this is reflected by all
        #   names (links) for the file.

        if($POSTSCRIPT{$file}) {
            my $cmd = $POSTSCRIPT{$file};
            print qq(RUNNING: $cmd\n);
            run_cmd($cmd);
        }
    }

    return 1;
}


sub user_is_root {

    if($< == 0) { return 1; }
    return undef;
}

sub report_improper_service_definition {

    my $name = shift;
    
    print STDERR "\n";
    print STDERR "Improper [service] definition.  Here's what I know about it:\n";
    print STDERR "  name   = $name\n";

    if(defined($DETAILS{$name})) { print STDERR "  mode   = $DETAILS{$name}\n";
                          } else { print STDERR "  mode   =\n"; }

    if(defined($DEPENDS{$name})) { print STDERR "  depends   = $DEPENDS{$name}\n";
                          } else { print STDERR "  depends   =\n"; }

    print STDERR "\n";
    print STDERR "  Skipping entry and incrementing ERROR_LEVEL...\n";
    print STDERR "\n";

    $ERROR_LEVEL++;

    return 1;
}


sub report_improper_file_definition {

    my $file = shift;
    
    print STDERR "\n";
    print STDERR "Improper [file] definition.  Here's what I know about it:\n";
    print STDERR "  name   = $file\n";

    if(defined($TYPE{$file})) { print STDERR "  type   = $TYPE{$file}\n";
                         } else { print STDERR "  type   =\n"; }

    if(defined($TARGET{$file})) { print STDERR "  target = $TARGET{$file}\n";
                           } else { print STDERR "  target =\n"; }

    if(defined($MODE{$file})) { print STDERR "  mode   = $MODE{$file}\n";
                         } else { print STDERR "  mode   =\n"; }

    if(defined($OWNER{$file})) { print STDERR "  owner  = $OWNER{$file}\n";
                          } else { print STDERR "  owner  =\n"; }

    if(defined($GROUP{$file})) { print STDERR "  group  = $GROUP{$file}\n";
                          } else { print STDERR "  group  =\n"; }

    if(defined($MAJOR{$file})) { print STDERR "  major  = $MAJOR{$file}\n";
                          } else { print STDERR "  major  =\n"; }

    if(defined($MINOR{$file})) { print STDERR "  minor  = $MINOR{$file}\n";
                          } else { print STDERR "  minor  =\n"; }

    if(defined($MD5SUM{$file})) { print STDERR "  md5sum = $MD5SUM{$file}\n";
                           } else { print STDERR "  md5sum =\n"; }

    if(defined($PRESCRIPT{$file})) { print STDERR "  prescript = $PRESCRIPT{$file}\n";
                              } else { print STDERR "  prescript =\n"; }

    if(defined($POSTSCRIPT{$file})) { print STDERR "  postscript = $POSTSCRIPT{$file}\n";
                               } else { print STDERR "  postscript =\n"; }

    if(defined($DEPENDS{$file})) { print STDERR "  depends = $DEPENDS{$file}\n";
                          } else { print STDERR "  depends =\n"; }

    if( defined($TYPE{$file}) and ($TYPE{$file} eq 'softlink') ) { 
        print STDERR "\n";
        print STDERR "  Make sure that file and target are absolute paths.  Relative\n";
        print STDERR "  links will still be created.\n";
    }

    print STDERR "\n";
    print STDERR "  Skipping entry and incrementing ERROR_LEVEL...\n";
    print STDERR "\n";

    $ERROR_LEVEL++;

    return 1;
}


sub uid_gid_and_mode_match {
    
    my $file = shift;

    my $st = stat($file);

    my $mode  = sprintf "%04o", $st_mode & 07777;

    if(    (   $mode == $MODE{$file}  )
        and( $st_uid == $OWNER{$file} )
        and( $st_gid == $GROUP{$file} ) ) {

        return 1;
    }

    return undef;
}

#
#   add_file_to_repo(@ARGV);
#
sub add_file_to_repo {

    my @chunks = @_;

    ####################################################################
    #
    # Validate input
    #
    if( ! -d $main::o{ou_path} ) {
        print "\n";
        print "The OU directory you specified does not exist:\n";
        print "  $main::o{ou_path}\n";
        print "\n";
        print "  Please create it first, set access privileges for it,\n";
        print "  then try again.\n";
        print "\n";
        exit 1;
    }

    foreach(@chunks) {
        if( ! -e $_ ) {
            print "CHUNK $_ doesn't appear to exist.\n";
            exit 1;
        }
    }
    #
    ####################################################################

    # Create the base dir
    my $dir = $main::o{ou_path} . '/' . $main::o{file};
    eval { mkpath($dir) };
    if($@) {
        print "Couldn’t create $dir: $@";
    }

    my $tmp_file = choose_tmp_file();
    print "tmp_file: $tmp_file\n";
    open(TMP, "+>$tmp_file") or die "Couldn't open tmp file $!";

        print "Concatenating files...\n";
        foreach my $file (@chunks) {
            print "CHUNK: $file\n" if($main::o{debug});
            open(FILE,"<$file") or die "Couldn't open $file for reading $!";
                print TMP <FILE>;
            close(FILE);
        }

        print "Calculating md5sum...\n";
        seek(TMP, 0, 0);
        my $md5sum = Digest::MD5->new->addfile(*TMP)->hexdigest;

        my $file = $dir . '/' . $md5sum;
        $file =~ s#/+#/#g;

    close(TMP);

    print "Moving target file into place:\n";
    print "  $file\n";
    move($tmp_file, $file) or die "Couldn't move($tmp_file, $file) $!";
    chmod oct(644), $file;

    $main::o{comment} = localtime() . " | " . $main::o{comment};

    $file = $file . ".README";
    print "Adding comments to README file:\n";
    print "  $file\n";
    open(FILE,">$file") or die "Couldn't open $file for writing $!";
        print FILE "COMMENT: $main::o{comment}\n";
        print FILE "\n"; 
        foreach(@chunks) {
            print FILE "CHUNK: $_\n";
        }
        print FILE "\n"; 
    close(FILE);
    chmod oct(644), $file;

    my $st = stat($chunks[0]);
    my $mode  = sprintf "%04o", $st_mode & 07777;
    my $owner = (getpwuid $st_uid)[0];
    my $group = (getgrgid $st_gid)[0];

    print "\n";
    print "Here's an entry you can cut and paste into your state definition\n";
    print "file.  Be sure to verify the owner, group, and mode values\n";
    print "------------------------------------------------------------------------\n";
    print "\n";
    print "[file]\n";
    print "comment  = $main::o{comment}\n";
    print "name     = $main::o{file}\n";
    print "type     = regular\n";
    print "md5sum   = $md5sum\n";
    print "owner    = $owner\n";
    print "group    = $group\n";
    print "mode     = $mode\n";
    print "priority = 0\n";
    print "\n";

    return 1;
}


sub turn_service_into_file_entry {

    my $name = shift;

    my $dir = "/etc";

    my %details;
    foreach( split(/\s+/, $DETAILS{$name}) ) {
        my ($level, $prefix) = split(/:/);
        $details{$level} = ${prefix};
    }

    opendir(DIR,"$dir");
    my @dirs = grep { /^rc.\.d$/ && -d "$dir/$_" } readdir(DIR);
    closedir(DIR);

    foreach my $subdir (@dirs) {

        #
        # Get a list of all links to the init script, and pre-determine
        # them as 'unwanted'.  Their state will be overridden further
        # below if they exist in the state definition.
        #
        opendir(DIR,"$dir/$subdir");
        foreach( grep { /[SK]\d+$name$/ } readdir(DIR) ) {
            my $file = "$dir/$subdir/$_";
            $TYPE{$file} = 'unwanted';
        }
        closedir(DIR);

        $subdir =~ m/rc(.)\.d/;
        my $level = $1;

        if(defined ($details{$level})) {
            my $prefix = $details{$level};
            my $file = "$dir/rc${level}.d/${prefix}${name}";
            $TYPE{$file} = 'softlink';
            $TARGET{$file} = "$dir/init.d/$name";
        }
    }

    return 1;
}


sub check_depends {

    my $name = shift;
    if(! defined $TYPE{$name}) {
        print ">> name: $name\n" if( $main::o{debug} );
        return 1;
    }

    my $unsatisfied;

    my %pkgs_currently_installed = get_pkgs_currently_installed();

    unless( "$TYPE{$name}" eq 'directory' ) {
        # add file's directory as dependency
        my $dirname = dirname $name;
        $DEPENDS{$name} .= " $dirname";
    };

    #
    # No dependencies to check?  That's OK.  Return success.
    if(! defined $DEPENDS{$name}) { return 1; }

    $DEPENDS{$name} =~ s/^\s+//;
    foreach( split(/\s+/, $DEPENDS{$name}) ) {
        if( /^\// ) {
            #
            # Must be a file.  If it doesn't exist, fail dep check.
            if( ! -e $_) { $unsatisfied .= "$_ "; }

        } else {
            #
            # Must be a package.  See if it's installed.
            if( ! defined $pkgs_currently_installed{$_} ) { $unsatisfied .= "$_ "; }
        }
    }

    if(defined $unsatisfied) { return $unsatisfied; }

    return 1;
}


#
# do we check for recursion loops?  Nah.  Not yet, anyway. ;-) -BEF-
#
sub _include_bundle {

    my $file = shift;

    $_ = $file;
    chomp($_);
    print "Including bundle:  $_\n";

    unless(($file =~ m#^file://#) 
        or ($file =~ m#^/#)
        or ($file =~ m#^http://#) 
        or ($file =~ m#^https://#) 
        or ($file =~ m#^ftp://#)) {

        $file = $main::o{base_uri} . '/' . $file;
    }

    my $tmp_file = get_file($file);

    open(FILE,"<$tmp_file") or die "Couldn't open $tmp_file for reading: $!";
        my @array = (<FILE>);
    close(FILE);
    unlink $tmp_file;

    #
    # Now add a blank line at the end of the array... -BEF-
    push( @array, "\n" );

    return @array;
}

sub _upgrade_ssm {

    my $pkg = 'ssm';

    my @pkgs_to_be_upgraded = get_pkgs_to_be_upgraded();
    foreach(@pkgs_to_be_upgraded) {

        if($_ eq $pkg) {

            print "FIXING: Upgrading $pkg\n";
            install_pkgs($pkg);

            my $cmd = $main::o{invocation_command};
            print "Restarting ssm with: $cmd\n";
            exec($cmd) or die("Couldn't exec $cmd");
        }
    }
    return 1;
}

sub _backup {

    my $file = shift;

    if( ! -e "/usr/bin/bu" ) { return 1; }
    if( ! -e $file ) { return 1; }

    my $cmd = "bu $file";
    !system($cmd) or die("FAILED: $cmd\n $!");

    return 1;
}


sub _add_file {

    my $file = shift;

    if( ! defined $main::o{svn_uri}) {
        print "ERROR: svn_uri not specified in definition file.  Please take a moment\n";
        print "       to add an entry to your [global] section.  Here's an example:\n";
        print "\n";
        print "         svn_uri = https://svn.example.com/repos/cis_ssm/trunk\n";
        print "\n";
        $ERROR_LEVEL++;
        exit $ERROR_LEVEL;
    }

    my $ou_path = "/tmp/ssm_db.repo.$$";
    my $svn     = qq(svn --no-auth-cache);

    if( ! -d "$ou_path/.ssm") {
        my $cmd = qq($svn checkout $main::o{svn_uri}/ $ou_path/);
        !system($cmd) or die("Couldn't run: $cmd");
    }

    my $hostname = `hostname -f`;
    chomp $hostname;
    my $comment = "$file on $hostname";
    my $cmd = qq(ssm_add-file --ou-path "$ou_path" --comment "$comment" --file "$file" "$file");
    !system($cmd) or die("Couldn't run: $cmd");

    chdir "$ou_path";
    $cmd = qq($svn stat);
    open(INPUT, "$cmd|") or die $!;
        while(<INPUT>) {
            chomp;
            s/\S+\s+//;

            chdir "$ou_path";
            $cmd = qq(svn add "$_");
            print qq($cmd\n);
            !system($cmd) or die("Couldn't run: $cmd");
            print "\n";
        }
    close(INPUT);

    $file =~ s#^/##;

    chdir "$ou_path";
    $cmd = qq($svn commit -m "$comment");
    print qq($cmd\n);
    !system($cmd) or die("Couldn't run: $cmd");
    print "\n";

    # Config chunks printed to the screen...
    print qq(Please take a moment to add the above config chunks to the appropriate\n);
    print qq(state definition file or bundle, and to commit your changes.  You may\n);
    print qq(want to open another terminal to do these things.\n);
    print qq(\n);
    print qq(Hit <Enter> to continue...\n);
    $_ = <STDIN>;

    return 1;
}

#
################################################################################

1;

